@@ -26,7 +26,7 @@
                                       /* ``Damned if you do,
                                            damned if you don't.''
                                                -- Unknown        */
-#include "mod_ssl.h"
+#include "ssl_private.h"
 
 /*  _________________________________________________________________
 **
@@ -72,7 +72,7 @@ SSLModConfigRec *ssl_config_global_create(server_rec *s)
     mc->tVHostKeys             = apr_hash_make(pool);
     mc->tPrivateKey            = apr_hash_make(pool);
     mc->tPublicCert            = apr_hash_make(pool);
-#ifdef SSL_EXPERIMENTAL_ENGINE
+#if defined(HAVE_OPENSSL_ENGINE_H) && defined(HAVE_ENGINE_INIT)
     mc->szCryptoDevice         = NULL;
 #endif
 
@@ -153,17 +153,9 @@ static void modssl_ctx_init_server(SSLSrvConfigRec *sc,
 
     modssl_ctx_init(mctx);
 
-    mctx->pks = apr_palloc(p, sizeof(*mctx->pks));
+    mctx->pks = apr_pcalloc(p, sizeof(*mctx->pks));
 
-    memset((void*)mctx->pks->cert_files, 0, sizeof(mctx->pks->cert_files));
-
-    memset((void*)mctx->pks->key_files, 0, sizeof(mctx->pks->key_files));
-
-    /* certs/keys are set during module init */
-
-    memset(mctx->pks->certs, 0, sizeof(mctx->pks->certs));
-
-    memset(mctx->pks->keys, 0, sizeof(mctx->pks->keys));
+    /* mctx->pks->... certs/keys are set during module init */
 }
 
 static SSLSrvConfigRec *ssl_config_server_new(apr_pool_t *p)
@@ -171,12 +163,21 @@ static SSLSrvConfigRec *ssl_config_server_new(apr_pool_t *p)
     SSLSrvConfigRec *sc = apr_palloc(p, sizeof(*sc));
 
     sc->mc                     = NULL;
-    sc->enabled                = UNSET;
+    sc->enabled                = SSL_ENABLED_FALSE;
     sc->proxy_enabled          = UNSET;
     sc->vhost_id               = NULL;  /* set during module init */
     sc->vhost_id_len           = 0;     /* set during module init */
     sc->session_cache_timeout  = UNSET;
+    sc->cipher_server_pref     = UNSET;
     sc->insecure_reneg         = UNSET;
+    sc->proxy_ssl_check_peer_expire = SSL_ENABLED_UNSET;
+    sc->proxy_ssl_check_peer_cn     = SSL_ENABLED_UNSET;
+#ifndef OPENSSL_NO_TLSEXT
+    sc->strict_sni_vhost_check = SSL_ENABLED_UNSET;
+#endif
+#ifdef HAVE_FIPS
+    sc->fips                   = UNSET;
+#endif
 
     modssl_ctx_init_proxy(sc, p);
 
@@ -246,6 +247,9 @@ static void modssl_ctx_cfg_merge_server(modssl_ctx_t *base,
         cfgMergeString(pks->cert_files[i]);
         cfgMergeString(pks->key_files[i]);
     }
+
+    cfgMergeString(pks->ca_name_path);
+    cfgMergeString(pks->ca_name_file);
 }
 
 /*
@@ -258,10 +262,19 @@ void *ssl_config_server_merge(apr_pool_t *p, void *basev, void *addv)
     SSLSrvConfigRec *mrg  = ssl_config_server_new(p);
 
     cfgMerge(mc, NULL);
-    cfgMergeBool(enabled);
+    cfgMerge(enabled, SSL_ENABLED_UNSET);
     cfgMergeBool(proxy_enabled);
     cfgMergeInt(session_cache_timeout);
+    cfgMergeBool(cipher_server_pref);
     cfgMergeBool(insecure_reneg);
+    cfgMerge(proxy_ssl_check_peer_expire, SSL_ENABLED_UNSET);
+    cfgMerge(proxy_ssl_check_peer_cn, SSL_ENABLED_UNSET);
+#ifndef OPENSSL_NO_TLSEXT
+    cfgMerge(strict_sni_vhost_check, SSL_ENABLED_UNSET);
+#endif
+#ifdef HAVE_FIPS
+    cfgMergeBool(fips);
+#endif
 
     modssl_ctx_cfg_merge_proxy(base->proxy, add->proxy, mrg->proxy);
 
@@ -291,6 +304,8 @@ void *ssl_config_perdir_create(apr_pool_t *p, char *dir)
     dc->szCACertificateFile    = NULL;
     dc->szUserName             = NULL;
 
+    dc->nRenegBufferSize = UNSET;
+
     return dc;
 }
 
@@ -328,6 +343,8 @@ void *ssl_config_perdir_merge(apr_pool_t *p, void *basev, void *addv)
     cfgMergeString(szCACertificateFile);
     cfgMergeString(szUserName);
 
+    cfgMergeInt(nRenegBufferSize);
+
     return mrg;
 }
 
@@ -337,10 +354,19 @@ void *ssl_config_perdir_merge(apr_pool_t *p, void *basev, void *addv)
 
 const char *ssl_cmd_SSLMutex(cmd_parms *cmd,
                              void *dcfg,
-                             const char *arg)
+                             const char *arg_)
 {
     const char *err;
     SSLModConfigRec *mc = myModConfig(cmd->server);
+    /* Split arg_ into meth and file */
+    char *meth = apr_pstrdup(cmd->temp_pool, arg_);
+    char *file = strchr(meth, ':');
+    if (file) {
+        *(file++) = '\0';
+        if (!*file) {
+            file = NULL;
+        }
+    }
 
     if ((err = ap_check_cmd_context(cmd, GLOBAL_ONLY))) {
         return err;
@@ -349,97 +375,71 @@ const char *ssl_cmd_SSLMutex(cmd_parms *cmd,
     if (ssl_config_global_isfixed(mc)) {
         return NULL;
     }
-
-    if (strcEQ(arg, "none") || strcEQ(arg, "no")) {
+    if (!strcasecmp(meth, "none") || !strcasecmp(meth, "no")) {
         mc->nMutexMode  = SSL_MUTEXMODE_NONE;
+        return NULL;
     }
+
+    /* APR determines temporary filename unless overridden below,
+     * we presume file indicates an szMutexFile is a file path
+     * unless the method sets szMutexFile=file and NULLs file
+     */
+    mc->nMutexMode  = SSL_MUTEXMODE_USED;
+    mc->szMutexFile = NULL;
+
     /* NOTE: previously, 'yes' implied 'sem' */
-    else if (strcEQ(arg, "default") || strcEQ(arg, "yes")) {
-        mc->nMutexMode  = SSL_MUTEXMODE_USED;
+    if (!strcasecmp(meth, "default") || !strcasecmp(meth, "yes")) {
         mc->nMutexMech = APR_LOCK_DEFAULT;
-        mc->szMutexFile = NULL; /* APR determines temporary filename */
     }
-#if APR_HAS_FLOCK_SERIALIZE
-    else if (strlen(arg) > 6 && strcEQn(arg, "flock:", 6)) {
-        const char *file = ap_server_root_relative(cmd->pool, arg+6);
-        if (!file) {
-            return apr_pstrcat(cmd->pool, "Invalid SSLMutex flock: path ", 
-                               arg+6, NULL);
-        }
-        mc->nMutexMode  = SSL_MUTEXMODE_USED;
-        mc->nMutexMech = APR_LOCK_FLOCK;
-        mc->szMutexFile = apr_psprintf(mc->pPool, "%s.%lu",
-                                       file, (unsigned long)getpid());
-    }
-#endif
 #if APR_HAS_FCNTL_SERIALIZE
-    else if (strlen(arg) > 6 && strcEQn(arg, "fcntl:", 6)) {
-        const char *file = ap_server_root_relative(cmd->pool, arg+6);
-        if (!file) {
-            return apr_pstrcat(cmd->pool, "Invalid SSLMutex fcntl: path ", 
-                               arg+6, NULL);
-        }
-        mc->nMutexMode  = SSL_MUTEXMODE_USED;
+    else if ((!strcasecmp(meth, "fcntl") || !strcasecmp(meth, "file")) && file) {
         mc->nMutexMech = APR_LOCK_FCNTL;
-        mc->szMutexFile = apr_psprintf(mc->pPool, "%s.%lu",
-                                       file, (unsigned long)getpid());
     }
 #endif
-#if APR_HAS_SYSVSEM_SERIALIZE && !defined(PERCHILD_MPM)
-    else if (strcEQ(arg, "sysvsem")) {
-        mc->nMutexMode  = SSL_MUTEXMODE_USED;
-        mc->nMutexMech = APR_LOCK_SYSVSEM;
-        mc->szMutexFile = NULL; /* APR determines temporary filename */
+#if APR_HAS_FLOCK_SERIALIZE
+    else if ((!strcasecmp(meth, "flock") || !strcasecmp(meth, "file")) && file) {
+        mc->nMutexMech = APR_LOCK_FLOCK;
     }
 #endif
 #if APR_HAS_POSIXSEM_SERIALIZE
-    else if (strcEQ(arg, "posixsem")) {
-        mc->nMutexMode  = SSL_MUTEXMODE_USED;
+    else if (!strcasecmp(meth, "posixsem") || !strcasecmp(meth, "sem")) {
         mc->nMutexMech = APR_LOCK_POSIXSEM;
-        mc->szMutexFile = NULL; /* APR determines temporary filename */
+        /* Posix/SysV semaphores aren't file based, use the literal name
+         * if provided and fall back on APR's default if not.  Today, APR
+         * will ignore it, but once supported it has an absurdly short limit.
+         */
+        if (file) {
+            mc->szMutexFile = apr_pstrdup(cmd->server->process->pool, file);
+
+            file = NULL;
+        }
+    }
+#endif
+#if APR_HAS_SYSVSEM_SERIALIZE && !defined(PERCHILD_MPM)
+    else if (!strcasecmp(meth, "sysvsem") || !strcasecmp(meth, "sem")) {
+        mc->nMutexMech = APR_LOCK_SYSVSEM;
     }
 #endif
 #if APR_HAS_PROC_PTHREAD_SERIALIZE
-    else if (strcEQ(arg, "pthread")) {
-        mc->nMutexMode  = SSL_MUTEXMODE_USED;
+    else if (!strcasecmp(meth, "pthread")) {
         mc->nMutexMech = APR_LOCK_PROC_PTHREAD;
-        mc->szMutexFile = NULL; /* APR determines temporary filename */
-    }
-#endif
-#if APR_HAS_FLOCK_SERIALIZE || APR_HAS_FCNTL_SERIALIZE
-    else if (strlen(arg) > 5 && strcEQn(arg, "file:", 5)) {
-        const char *file = ap_server_root_relative(cmd->pool, arg+5);
-        if (!file) {
-            return apr_pstrcat(cmd->pool, "Invalid SSLMutex file: path ", 
-                               arg+5, NULL);
-        }
-        mc->nMutexMode  = SSL_MUTEXMODE_USED;
-#if APR_HAS_FLOCK_SERIALIZE
-        mc->nMutexMech  = APR_LOCK_FLOCK;
-#endif
-#if APR_HAS_FCNTL_SERIALIZE
-        mc->nMutexMech  = APR_LOCK_FCNTL;
-#endif
-        mc->szMutexFile =
-            apr_psprintf(mc->pPool, "%s.%lu",
-                         file, (unsigned long)getpid());
-    }
-#endif
-#if (APR_HAS_SYSVSEM_SERIALIZE && !defined(PERCHILD_MPM)) || APR_HAS_POSIXSEM_SERIALIZE
-    else if (strcEQ(arg, "sem")) {
-        mc->nMutexMode  = SSL_MUTEXMODE_USED;
-#if APR_HAS_SYSVSEM_SERIALIZE && !defined(PERCHILD_MPM)
-        mc->nMutexMech  = APR_LOCK_SYSVSEM;
-#endif
-#if APR_HAS_POSIXSEM_SERIALIZE
-        mc->nMutexMech  = APR_LOCK_POSIXSEM;
-#endif
-        mc->szMutexFile = NULL; /* APR determines temporary filename */
     }
 #endif
     else {
-        return apr_pstrcat(cmd->pool, "Invalid SSLMutex argument ", 
-                           arg, " (", ssl_valid_ssl_mutex_string, ")", NULL);
+        return apr_pstrcat(cmd->pool, "Invalid SSLMutex argument ", arg_,
+                           " (", ssl_valid_ssl_mutex_string, ")", NULL);
+    }
+
+    /* Unless the method above assumed responsibility for setting up
+     * mc->szMutexFile and NULLing out file, presume it is a file we
+     * are looking to use
+     */
+    if (file) {
+        mc->szMutexFile = ap_server_root_relative(cmd->server->process->pool, file);
+        if (!mc->szMutexFile) {
+            return apr_pstrcat(cmd->pool, "Invalid SSLMutex ", meth,
+                               ": filepath ", file, NULL);
+        }
     }
 
     return NULL;
@@ -463,7 +463,6 @@ const char *ssl_cmd_SSLPassPhraseDialog(cmd_parms *cmd,
     }
     else if ((arglen > 5) && strEQn(arg, "exec:", 5)) {
         sc->server->pphrase_dialog_type  = SSL_PPTYPE_FILTER;
-        /* ### This is broken, exec: may contain args, no? */
         sc->server->pphrase_dialog_path =
             ap_server_root_relative(cmd->pool, arg+5);
         if (!sc->server->pphrase_dialog_path) {
@@ -493,7 +492,7 @@ const char *ssl_cmd_SSLPassPhraseDialog(cmd_parms *cmd,
     return NULL;
 }
 
-#ifdef SSL_EXPERIMENTAL_ENGINE
+#if defined(HAVE_OPENSSL_ENGINE_H) && defined(HAVE_ENGINE_INIT)
 const char *ssl_cmd_SSLCryptoDevice(cmd_parms *cmd,
                                     void *dcfg,
                                     const char *arg)
@@ -501,16 +500,7 @@ const char *ssl_cmd_SSLCryptoDevice(cmd_parms *cmd,
     SSLModConfigRec *mc = myModConfig(cmd->server);
     const char *err;
     ENGINE *e;
-#if SSL_LIBRARY_VERSION >= 0x00907000
-    static int loaded_engines = FALSE;
 
-    /* early loading to make sure the engines are already 
-       available for ENGINE_by_id() above... */
-    if (!loaded_engines) {
-        ENGINE_load_builtin_engines();
-        loaded_engines = TRUE;
-    }
-#endif
     if ((err = ap_check_cmd_context(cmd, GLOBAL_ONLY))) {
         return err;
     }
@@ -523,7 +513,18 @@ const char *ssl_cmd_SSLCryptoDevice(cmd_parms *cmd,
         ENGINE_free(e);
     }
     else {
-        return "SSLCryptoDevice: Invalid argument";
+        err = "SSLCryptoDevice: Invalid argument; must be one of: "
+              "'builtin' (none)";
+        e = ENGINE_get_first();
+        while (e) {
+            ENGINE *en;
+            err = apr_pstrcat(cmd->pool, err, ", '", ENGINE_get_id(e),
+                                         "' (", ENGINE_get_name(e), ")", NULL);
+            en = ENGINE_get_next(e);
+            ENGINE_free(e);
+            e = en;
+        }
+        return err;
     }
 
     return NULL;
@@ -532,7 +533,7 @@ const char *ssl_cmd_SSLCryptoDevice(cmd_parms *cmd,
 
 const char *ssl_cmd_SSLRandomSeed(cmd_parms *cmd,
                                   void *dcfg,
-                                  const char *arg1, 
+                                  const char *arg1,
                                   const char *arg2,
                                   const char *arg3)
 {
@@ -620,11 +621,45 @@ const char *ssl_cmd_SSLRandomSeed(cmd_parms *cmd,
     return NULL;
 }
 
-const char *ssl_cmd_SSLEngine(cmd_parms *cmd, void *dcfg, int flag)
+const char *ssl_cmd_SSLEngine(cmd_parms *cmd, void *dcfg, const char *arg)
 {
     SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
 
-    sc->enabled = flag ? TRUE : FALSE;
+    if (!strcasecmp(arg, "On")) {
+        sc->enabled = SSL_ENABLED_TRUE;
+    return NULL;
+    }
+    else if (!strcasecmp(arg, "Off")) {
+        sc->enabled = SSL_ENABLED_FALSE;
+        return NULL;
+    }
+    else if (!strcasecmp(arg, "Optional")) {
+        sc->enabled = SSL_ENABLED_OPTIONAL;
+        return NULL;
+    }
+
+    return "Argument must be On, Off, or Optional";
+}
+
+const char *ssl_cmd_SSLFIPS(cmd_parms *cmd, void *dcfg, int flag)
+{
+#ifdef HAVE_FIPS
+    SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
+#endif
+    const char *err;
+
+    if ((err = ap_check_cmd_context(cmd, GLOBAL_ONLY))) {
+        return err;
+    }
+
+#ifdef HAVE_FIPS
+    if ((sc->fips != UNSET) && (sc->fips != (BOOL)(flag ? TRUE : FALSE)))
+        return "Conflicting SSLFIPS options, cannot be both On and Off";
+    sc->fips = flag ? TRUE : FALSE;
+#else
+    if (flag)
+        return "SSLFIPS invalid, rebuild httpd and openssl compiled for FIPS";
+#endif
 
     return NULL;
 }
@@ -668,11 +703,22 @@ static const char *ssl_cmd_check_file(cmd_parms *parms,
     }
 
     return apr_pstrcat(parms->pool, parms->cmd->name,
-                       ": file '", *file, 
+                       ": file '", *file,
                        "' does not exist or is empty", NULL);
 
 }
 
+const char *ssl_cmd_SSLHonorCipherOrder(cmd_parms *cmd, void *dcfg, int flag)
+{
+#ifdef SSL_OP_CIPHER_SERVER_PREFERENCE
+    SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
+    sc->cipher_server_pref = flag?TRUE:FALSE;
+    return NULL;
+#else
+    return "SSLHonorCiperOrder unsupported; not implemented by the SSL library";
+#endif
+}
+
 const char *ssl_cmd_SSLInsecureRenegotiation(cmd_parms *cmd, void *dcfg, int flag)
 {
 #ifdef SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION
@@ -702,7 +748,7 @@ static const char *ssl_cmd_check_dir(cmd_parms *parms,
     }
 
     return apr_pstrcat(parms->pool, parms->cmd->name,
-                       ": directory '", *dir, 
+                       ": directory '", *dir,
                        "' does not exist", NULL);
 
 }
@@ -742,7 +788,7 @@ static const char *ssl_cmd_check_aidx_max(cmd_parms *parms,
 
     return apr_psprintf(parms->pool,
                         "%s: only up to %d "
-                        "different %s per virtual host allowed", 
+                        "different %s per virtual host allowed",
                          parms->cmd->name, SSL_AIDX_MAX, desc);
 }
 
@@ -843,6 +889,36 @@ const char *ssl_cmd_SSLCACertificateFile(cmd_parms *cmd,
     return NULL;
 }
 
+const char *ssl_cmd_SSLCADNRequestPath(cmd_parms *cmd, void *dcfg,
+                                       const char *arg)
+{
+    SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
+    const char *err;
+
+    if ((err = ssl_cmd_check_dir(cmd, &arg))) {
+        return err;
+    }
+
+    sc->server->pks->ca_name_path = arg;
+
+    return NULL;
+}
+
+const char *ssl_cmd_SSLCADNRequestFile(cmd_parms *cmd, void *dcfg,
+                                       const char *arg)
+{
+    SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
+    const char *err;
+
+    if ((err = ssl_cmd_check_file(cmd, &arg))) {
+        return err;
+    }
+
+    sc->server->pks->ca_name_file = arg;
+
+    return NULL;
+}
+
 const char *ssl_cmd_SSLCARevocationPath(cmd_parms *cmd,
                                         void *dcfg,
                                         const char *arg)
@@ -912,7 +988,7 @@ const char *ssl_cmd_SSLVerifyClient(cmd_parms *cmd,
     if ((err = ssl_cmd_verify_parse(cmd, arg, &mode))) {
         return err;
     }
-    
+
     if (cmd->path) {
         dc->nVerifyClient = mode;
     }
@@ -983,6 +1059,10 @@ const char *ssl_cmd_SSLSessionCache(cmd_parms *cmd,
         mc->nSessionCacheMode      = SSL_SCMODE_NONE;
         mc->szSessionCacheDataFile = NULL;
     }
+    else if (strcEQ(arg, "nonenotnull")) {
+        mc->nSessionCacheMode      = SSL_SCMODE_NONE_NOT_NULL;
+        mc->szSessionCacheDataFile = NULL;
+    }
     else if ((arglen > 4) && strcEQn(arg, "dbm:", 4)) {
         mc->nSessionCacheMode      = SSL_SCMODE_DBM;
         mc->szSessionCacheDataFile = ap_server_root_relative(mc->pPool, arg+4);
@@ -992,48 +1072,8 @@ const char *ssl_cmd_SSLSessionCache(cmd_parms *cmd,
                                 arg+4);
         }
     }
-    else if ((arglen > 6) && strcEQn(arg, "shmht:", 6)) {
-#if !APR_HAS_SHARED_MEMORY
-        return MODSSL_NO_SHARED_MEMORY_ERROR;
-#endif
-        mc->nSessionCacheMode = SSL_SCMODE_SHMHT;
-        colon = ap_strchr_c(arg, ':');
-        mc->szSessionCacheDataFile =
-            ap_server_root_relative(mc->pPool, colon+1);
-        if (!mc->szSessionCacheDataFile) {
-            return apr_psprintf(cmd->pool,
-                                "SSLSessionCache: Invalid cache file path %s",
-                                colon+1);
-        }
-        mc->tSessionCacheDataTable = NULL;
-        mc->nSessionCacheDataSize  = 1024*512; /* 512KB */
-
-        if ((cp = strchr(mc->szSessionCacheDataFile, '('))) {
-            *cp++ = NUL;
-
-            if (!(cp2 = strchr(cp, ')'))) {
-                return "SSLSessionCache: Invalid argument: "
-                       "no closing parenthesis";
-            }
-
-            *cp2 = NUL;
-
-            mc->nSessionCacheDataSize = atoi(cp);
-
-            if (mc->nSessionCacheDataSize < 8192) {
-                return "SSLSessionCache: Invalid argument: "
-                       "size has to be >= 8192 bytes";
-            }
-
-            if (mc->nSessionCacheDataSize >= APR_SHM_MAXSIZE) {
-                return apr_psprintf(cmd->pool,
-                                    "SSLSessionCache: Invalid argument: "
-                                    "size has to be < %d bytes on this "
-                                    "platform", APR_SHM_MAXSIZE);
-            }
-        }
-    }
     else if (((arglen > 4) && strcEQn(arg, "shm:", 4)) ||
+             ((arglen > 6) && strcEQn(arg, "shmht:", 6)) ||
              ((arglen > 6) && strcEQn(arg, "shmcb:", 6))) {
 #if !APR_HAS_SHARED_MEMORY
         return MODSSL_NO_SHARED_MEMORY_ERROR;
@@ -1077,6 +1117,19 @@ const char *ssl_cmd_SSLSessionCache(cmd_parms *cmd,
             }
         }
     }
+    else if ((arglen > 3) && strcEQn(arg, "dc:", 3)) {
+#ifdef HAVE_DISTCACHE
+        mc->nSessionCacheMode      = SSL_SCMODE_DC;
+        mc->szSessionCacheDataFile = apr_pstrdup(mc->pPool, arg+3);
+        if (!mc->szSessionCacheDataFile) {
+            return apr_pstrcat(cmd->pool,
+                               "SSLSessionCache: Invalid cache file path: ",
+                               arg+3, NULL);
+        }
+#else
+        return "SSLSessionCache: distcache support disabled";
+#endif
+    }
     else {
         return "SSLSessionCache: Invalid argument";
     }
@@ -1123,9 +1176,6 @@ const char *ssl_cmd_SSLOptions(cmd_parms *cmd,
         if (strcEQ(w, "StdEnvVars")) {
             opt = SSL_OPT_STDENVVARS;
         }
-        else if (strcEQ(w, "CompatEnvVars")) {
-            opt = SSL_OPT_COMPATENVVARS;
-        }
         else if (strcEQ(w, "ExportCertData")) {
             opt = SSL_OPT_EXPORTCERTDATA;
         }
@@ -1193,6 +1243,19 @@ const char *ssl_cmd_SSLRequire(cmd_parms *cmd,
     return NULL;
 }
 
+const char *ssl_cmd_SSLRenegBufferSize(cmd_parms *cmd, void *dcfg, const char *arg)
+{
+    SSLDirConfigRec *dc = dcfg;
+    
+    dc->nRenegBufferSize = atoi(arg);
+    if (dc->nRenegBufferSize < 0) {
+        return apr_pstrcat(cmd->pool, "Invalid size for SSLRenegBufferSize: ",
+                           arg, NULL);
+    }
+
+    return NULL;
+}
+
 static const char *ssl_cmd_protocol_parse(cmd_parms *parms,
                                           const char *arg,
                                           ssl_proto_t *options)
@@ -1260,7 +1323,7 @@ const char *ssl_cmd_SSLProxyEngine(cmd_parms *cmd, void *dcfg, int flag)
     return NULL;
 }
 
-const char *ssl_cmd_SSLProxyProtocol(cmd_parms *cmd, 
+const char *ssl_cmd_SSLProxyProtocol(cmd_parms *cmd,
                                      void *dcfg,
                                      const char *arg)
 {
@@ -1411,10 +1474,68 @@ const char *ssl_cmd_SSLProxyMachineCertificatePath(cmd_parms *cmd,
 }
 
 
-const char *ssl_cmd_SSLUserName(cmd_parms *cmd, void *dcfg, 
-				const char *arg)
+const char *ssl_cmd_SSLUserName(cmd_parms *cmd, void *dcfg,
+                                const char *arg)
 {
     SSLDirConfigRec *dc = (SSLDirConfigRec *)dcfg;
     dc->szUserName = arg;
     return NULL;
 }
+
+const char *ssl_cmd_SSLProxyCheckPeerExpire(cmd_parms *cmd, void *dcfg, int flag)
+{
+    SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
+
+    sc->proxy_ssl_check_peer_expire = flag ? SSL_ENABLED_TRUE : SSL_ENABLED_FALSE;
+
+    return NULL;
+}
+
+const char *ssl_cmd_SSLProxyCheckPeerCN(cmd_parms *cmd, void *dcfg, int flag)
+{
+    SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
+
+    sc->proxy_ssl_check_peer_cn = flag ? SSL_ENABLED_TRUE : SSL_ENABLED_FALSE;
+
+    return NULL;
+}
+
+const char  *ssl_cmd_SSLStrictSNIVHostCheck(cmd_parms *cmd, void *dcfg, int flag)
+{
+#ifndef OPENSSL_NO_TLSEXT
+    SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
+
+    sc->strict_sni_vhost_check = flag ? SSL_ENABLED_TRUE : SSL_ENABLED_FALSE;
+
+    return NULL;
+#else
+    return "SSLStrictSNIVHostCheck failed; OpenSSL is not built with support "
+           "for TLS extensions and SNI indication. Refer to the "
+           "documentation, and build a compatible version of OpenSSL.";
+#endif
+}
+
+void ssl_hook_ConfigTest(apr_pool_t *pconf, server_rec *s)
+{
+    if (!ap_exists_config_define("DUMP_CERTS")) {
+        return;
+    }
+
+    /* Dump the filenames of all configured server certificates to
+     * stdout. */
+    while (s) {
+        SSLSrvConfigRec *sc = mySrvConfig(s);
+
+        if (sc && sc->server && sc->server->pks) {
+            modssl_pk_server_t *const pks = sc->server->pks;
+            int i;
+
+            for (i = 0; (i < SSL_AIDX_MAX) && pks->cert_files[i]; i++) {
+                printf("%s\n", pks->cert_files[i]);
+            }
+        }
+
+        s = s->next;
+    }
+
+}
